package main

import (
	"flag"
	"fmt"
	"git.gammaspectra.live/S.O.N.G/MeteorLight/util"
	"log"
	"net/http"
	"sync"
	"time"
)

func main() {
	configPath := flag.String("config", "config.toml", "Config path")

	flag.Parse()

	config, err := GetConfig(*configPath)
	if err != nil {
		log.Panic(err)
	}

	var wg sync.WaitGroup

	queue := NewQueue(config)

	wg.Add(1)
	go func() {
		defer wg.Done()

		server := http.Server{
			Addr:    fmt.Sprintf("%s:%d", config.Radio.Host, config.Radio.Port),
			Handler: http.HandlerFunc(queue.HandleRadioRequest),
		}
		server.SetKeepAlivesEnabled(false)

		//setup a timeout to prevent slow clients blocking. See https://github.com/golang/go/issues/16100
		timeoutListener, err := util.NewTimeoutListener("tcp", server.Addr, time.Second*15, time.Second*15)
		if err != nil {
			log.Panic(err)
		}

		if err = server.Serve(timeoutListener); err != nil {
			log.Panic(err)
		}
	}()

	api := NewAPI(config, queue)

	api.Wait()
	wg.Wait()
	queue.Wait()
}
